(require-library lazy-lists clojurian-syntax)
(import lazy-lists clojurian-syntax)

;;; (run xpr0 xpr1 ...)
;;; -------------------
(define (run . xprs)
  (let loop ((xprs xprs))
    (if (null? xprs)
      (print "All tests passed!")
      (if (car xprs)
        (loop (cdr xprs))
        (error 'run "#### Some test failed! ####")))))

(define (cons-right var lst)
  (if (null? lst)
    (cons var lst)
    (cons (car lst) (cons-right var (cdr lst)))))

(define (Within eps seq)
  (let loop ((seq seq))
    (let ((a (Ref 0 seq)) (b (Ref 1 seq)))
      (if (< (abs (- a b)) eps)
        b
        (loop (Rest seq))))))

(define (Relative eps seq)
  (let loop ((seq seq))
    (let ((a (Ref 0 seq)) (b (Ref 1 seq)))
      (if (<= (abs (/ a b)) (* (abs b) eps))
        b
        (loop (Rest seq))))))

(define (Newton x) ; fixed point for square root
  (lambda (a) (/ (+ a (/ x a)) 2))) 

(define (Sums seq) ; List of sums
  (letrec ((sums (Cons 0 (Map + seq (Lazy (Length seq) sums)))))
    (Rest sums)))

(define (First-five) (List 0 1 2 3 4))
(define (Fibs)
  (Append (List 0 1) (Lazy #f (Map + (Rest (Fibs)) (Fibs)))))

;(define port (open-input-file "lazy-lists.scm"))
;(define input (input->List port read-line))

(run
  (= (Length (First-five)) 5)
  (= (Length (Rest (First-five))) 4)
  (->> (Cardinals) (Rest) (Length) (eq? #f))
  (->> (Cardinals) (Take 5) (Length) (= 5))
  (->> (Cardinals) (Length) (eq? #f))
  (->> (Cardinals) (Drop 5) (Length) (eq? #f))
  (->> (Cardinals) (Drop 5) (First) (= 5))
  (->> (First-five) (List->list) (equal? '(0 1 2 3 4)))
  (->> (Cardinals) (Take 5) (List->list) (equal? '(0 1 2 3 4)))
  (->> (Interval 2 10) (Length) (= (- 10 2)))
  (->> (Interval 2 10) (List->list) (equal? '(2 3 4 5 6 7 8 9)))
  (->> (Interval 10 2) (List->list) (equal? '(10 9 8 7 6 5 4 3)))
  (equal?
    (receive (head index tail) (Split-with (cut = <> 3) (First-five))
      (cons (First tail) (List->list head)))
    '(3 0 1 2))
  (equal?
    (receive (head index tail) (Split-with (cut = <> 5) (Take 10 (Cardinals)))
      (append (List->list tail) (List->list head)))
    '(5 6 7 8 9 0 1 2 3 4))
  (->> (First-five) (Index (cut = <> 2)) (= 2))
  (->> (First-five) (Index (cut = <> 20)) (= 5))
  (->> (Cardinals) (Take 10) (Take-upto (cut = <> 5)) (List->list)
       (equal? '(0 1 2 3 4)))
  (->> (Cardinals) (Take 10) (Take-upto (cut = <> 5)) (Length) (= 5))
  (->> (Cardinals) (Take 10) (Drop-upto (cut = <> 5)) (Length) (= 5))
  (->> (Cardinals) (Take 10) (Drop-upto (cut = <> 5)) (First) (= 5))
  (->> (First-five) (Drop-upto (cut = <> 2)) (Length) (= 3))
  (->> (First-five) (Drop-upto (cut = <> 2)) (First) (= 2))
  (->> (First-five) (Memp odd?) (List->list) (equal? '(1 2 3 4)))
  (->> (Cardinals) (Take 10) (Memv 5) (List->list) (equal? '(5 6 7 8 9)))
  (->> (Cardinals) (Map (lambda (x) (list x x))) (Take 10) (Assv 5)
       (equal? '(5 5)))
  (->> (First-five) (Map (lambda (x) (list x x))) (Assv 10)
       (eq? #f))
  (->> (Cardinals) (Equal? (Cardinals)) (eq? #f))
  (->> (Cardinals) (Equal? (First-five)) (eq? #f))
  (->> (First-five) (Equal? (First-five)) (eq? #t))
  (->> (Cardinals) (Take 10) (Length) (= 10))
  (->> (Cardinals) (Drop 1) (Filter odd?) (Take 5) (List->list)
       (equal? '(1 3 5 7 9)))
  (->> (Cardinals) (Length) (eq? #f))
  (->> (First-five) (Map add1) (List->list) (equal? '(1 2 3 4 5)))
  (->> (Map + (First-five) (Take 5 (Cardinals))) (List->list)
       (equal? '(0 2 4 6 8)))
  (->> (Map + (Cardinals) (Cardinals)) (Length) (eq? #f))
  (->> (Filter odd? (First-five)) (Length) (= 2))
  (->> (Filter odd? (First-five)) (List->list) (equal? '(1 3)))
  (->> (Filter odd? (Cardinals)) (Length) (eq? #f))
  (->> (Zip (Cardinals) (Cardinals)) (Sieve =) (Ref 20) (= 20))
  (->> (Zip (First-five) (First-five)) (Sieve =) (List->list)
       (equal? '(0 1 2 3 4)))
  (->> (Ref 25 (Cardinals)) (= 25))
  (->> (Ref 2 (First-five)) (= 2))
  (->> (Repeat #f) (Take 3) (List->list) (equal? '(#f #f #f)))
  (->> (Repeatedly (lambda () 1))(Take 3)  (List->list)
       (equal? '(1 1 1)))
  (->> (Iterate add1 0) (Take 3) (List->list) (equal? '(0 1 2)))
  (->> (Iterate add1 0) (Length) (eq? #f))
  (->> (Append (First-five) (First-five)) (Length) (= 10))
  (->> (Append (First-five) (First-five)) (List->list)
       (equal? '(0 1 2 3 4 0 1 2 3 4)))
  (->> (Append (First-five) (Cardinals)) (Take 12) (List->list)
       (equal? '(0 1 2 3 4 0 1 2 3 4 5 6)))
  (->> (Append (First-five) (Cardinals)) (Length) (eq? #f))
  (->> (First-five) (Reverse) (List->list) (equal? '(4 3 2 1 0)))
  (->> (Cardinals) (Take 5) (Reverse) (List->list) (equal? '(4 3 2 1 0)))
  (->> (First-five) (Reverse) (Length) (= 5))
  (->> (Cardinals) (Reverse*) (Length) (eq? #f))
  (->> (Cardinals) (Reverse*) (Ref 5) (List->list) (equal? '(5 4 3 2 1 0)))
  (->> (Cycle (First-five)) (Take 10) (List->list)
       (equal? '(0 1 2 3 4 0 1 2 3 4)))
  (->> (Cycle (First-five)) (Length) (eq? #f))
  (->> (Sort < (First-five)) (List->list) (equal? '(0 1 2 3 4)))
  (->> (Sort < (First-five)) (Length) (= 5))
  (->> (Sort < (List 3 1 0 2 4)) (List->list) (equal? '(0 1 2 3 4)))
  (equal?
    (receive (head tail) (Split-at 5 (Cardinals))
      (cons (First tail) (List->list head)))
    '(5 0 1 2 3 4))
  (equal?
    (receive (head tail) (Split-at 15 (Take 5 (Cardinals)))
      (append (List->list tail) (List->list head)))
    '(0 1 2 3 4))
  (->> (Cardinals) (Take 5) (Fold-left + 0) (= 10))
  (->> (Fold-left + 0 (First-five) (First-five)) (= 20))
  (->> (List 1 2 3 4) (Fold-left * 1) (= 24))
  (->> (Cardinals) (Take 5) (Fold-left cons '())
       (equal? '(((((() . 0) . 1) . 2) . 3) . 4)))
  (->> (Cardinals) (Fold-left* cons '()) (Ref 4)
       (equal? '(((((() . 0) . 1) . 2) . 3) . 4)))
  (->> (Cardinals) (Take 5) (Fold-right + 0) (= 10))
  (->> (Fold-right + 0 (First-five) (First-five)) (= 20))
  (->> (First-five) (Fold-right cons '())
       (equal? '(0 1 2 3 4))) ; list
  (->> (First-five) (Fold-right cons '(a b c))
       (equal? '(0 1 2 3 4 a b c))) ; append
  (->> (Cardinals) (Fold-right* cons '()) (Ref 4)
       (equal? '(4 3 2 1 0))) ; note changed order
  (->> (Cardinals) (Fold-right* cons-right '()) (Ref 4)
       (equal? '(0 1 2 3 4)))
  (->> (Cardinals) (Fold-right* cons '(a b c)) (Ref 4)
       (equal? '(4 3 2 1 0 a b c))) ; note changed order
  (->> (Cardinals) (Fold-right* cons-right '(a b c)) (Ref 4)
       (equal? '(a b c 0 1 2 3 4)))
  (->> (vector->List '#(0 1 2 3 4)) (List->list) (equal? '(0 1 2 3 4)))
  (->> (vector->List '#()) (Null?))
  (->> (Take 5 (Cardinals)) (List->vector) (equal? '#(0 1 2 3 4)))
  (->> (List->vector (First-five)) (equal? '#(0 1 2 3 4)))
  (->> (List->vector Nil) (equal? '#()))
  (->> (Filter odd? (Cardinals)) (Take 15) (Every? odd?)
       (eq? #t))
  (->> (Take 15 (Cardinals)) (Every? odd?)
       (eq? #f))
  (->> (Every? odd? Nil) (eq? #t))
  (->> (Some? odd? Nil) (eq? #f))
  (->> (Filter even? (Cardinals)) (Take 5) (Some? odd?)
       (eq? #f))
  (->> (Some? odd? (First-five)) (eq? #t))
  (->> (Zip (Cardinals) (First-five)) (Length) (eq? #f))
  (->> (Zip (First-five) (Cardinals)) (Length) (eq? #f))
  (->> (Zip (Cardinals) (Cardinals)) (Length) (eq? #f))
  (->> (Zip (Cardinals) (First-five)) (Take 14)
       (Eqv? (List 0 0 1 1 2 2 3 3 4 4 5 6 7 8)))
  (->> (Zip (Cardinals) (Cardinals)) (Take 14)
       (Eqv? (List 0 0 1 1 2 2 3 3 4 4 5 5 6 6)))
  (->> (Zip (First-five) (First-five)) (Length) (= 10))
  (->> (Zip (First-five) (First-five))
       (Eqv? (List 0 0 1 1 2 2 3 3 4 4)))
  (->> (Primes) (Ref 50) (= 233))
  (->> (Primes) (Take 5) (Eqv? (List 2 3 5 7 11)))
  (->> (Fibs) (Take 10) (Eqv? (List  0 1 1 2 3 5 8 13 21 34)))
  ;; compute square root
  (let ((eps 0.000001))
    (< (abs (- (sqrt 2) (Within eps (Iterate (Newton 2) 2)))) eps))
  (->> (First-five) (Sums) (List->list) (equal? '(0 1 3 6 10)))
  )

